#!/usr/bin/env python
#       "$Name:  $";
#       "$Header:  $";
#=============================================================================
#
# file :        FlexHV.py
#
# description : Python source for the FlexHV Tango device
#               The class uses Socket Tango Device to communicate with the
#               Flex High Voltage Power Supply (made by ELETTRA)
#
# project :     TANGO Device Server
#
# $Author:  $ Elettra SCI.COMP.GROUP.
#
# $Revision:  $
#
#=============================================================================
#
import sys
import PyTango
import time

#==================================================================
#   Device States Description:
#
#   DevState.INIT :
#   DevState.ON : Standby/running
#   DevState.FAULT : Connection to Socket lost.
#==================================================================
class FlexHV(PyTango.Device_4Impl):

#--------- Add you global variables here --------------------------


#------------------------------------------------------------------
#    Device constructor
#------------------------------------------------------------------
    def __init__(self,cl, name):
        print 'In Device constructor'
        PyTango.Device_4Impl.__init__(self,cl,name)
        FlexHV.init_device(self)

#------------------------------------------------------------------
#    Device destructor
#------------------------------------------------------------------
    def delete_device(self):       
        print "[Device delete_device method] for device",self.get_name()

#------------------------------------------------------------------
#    Device initialization
#------------------------------------------------------------------
    def init_device(self):
        self.debug_stream("In " + self.get_name() + "::init_device()")
        self.set_state(PyTango.DevState.INIT)
        self.status_string = "Initializing device..."
        self.get_device_properties(self.get_device_class())

        self.display_txt = 'UNKNOWN'
        self.max_voltage = 0.0
        self.min_voltage = 0.0

        self.calib_offset = 0.0
        self.calib_slope  = 0.0
        self.calib_points = 0
        self.calib_table  = []

        try:
            # socket device to FLEX HV Power Supply
            self.socket = PyTango.DeviceProxy(self.SocketDevice)

            if self.check_socket():
                if self.read_maxV_from_device() \
                and self.initialize_to_offset() \
                and self.read_calibration_parameters():
                    self.set_status("Device initialized.")
                    self.debug_stream("Device initialized.")
                    self.set_state(PyTango.DevState.ON)
                else:
                    self.set_state(PyTango.DevState.FAULT)
                    self.set_status("INIT failed")
                    self.error_stream("INIT failed")
            else:
                self.set_state(PyTango.DevState.FAULT)
                self.set_status("INIT failed")
                self.error_stream("INIT failed")

        except PyTango.DevFailed:
            exctype , value = sys.exc_info()[:2]
            self.error_stream("Socket failed.")
            for err in value:
                self.error_stream(" reason" + str(err.reason))
                self.error_stream(" description" + str(err.desc))
                self.error_stream(" origin" + str(err.origin))
                self.error_stream(" severity" + str(err.severity))

            self.set_state(PyTango.DevState.FAULT)
            self.set_status("INIT failed, Socket proxy failed.")
            self.error_stream("INIT failed, Socket proxy failed.")

#------------------------------------------------------------------
#    Always excuted hook method
#------------------------------------------------------------------
    def always_executed_hook(self):
        # print "In ", self.get_name(), "::always_executed_hook()"
        self.set_status(self.status_string)

#==================================================================
#
#    FlexHV utility methods
#
#==================================================================

#------------------------------------------------------------------
#    Check socket status:
#    returns True if socket is in ON state, False otherwise
#------------------------------------------------------------------
    def check_socket(self):
        try:
            if self.socket.state() == PyTango.DevState.ON:
                self.status_string = "Socket device check OK, socket in state ON"
                return 1
            else:
               self.status_string = "Socket device failed, socket in state: " + self.socket.state()
               self.debug_stream(self.status_string)
        except Exception, e:
            self.status_string = "Connection failed, socket error: " + e
            self.debug_stream(self.status_string)
        return 0

#------------------------------------------------------------------
#    Read/write socket utility
#        Returns a string in case of a proper reply,
#        1 for AKC reply and 0 for NAK reply or failure.
#------------------------------------------------------------------
    def socket_write_and_read(self, argin):
        # print "In ", self.get_name(), "::socket_write_and_read()"

        # end_of_line = '\r\n' #apparently, already there
        resp_ok = 'ACK'
        resp_not_ok = 'NAK'

        # Flush the socket buffer
        self.socket.Read()

        argin = str(argin) # + end_of_line
        # self.debug_stream("argin: " + argin)
        attempts = 0
        max_attempts = 2 # In case of a socket exception, retry once.
        while (attempts < max_attempts):
            try:
                argout = self.socket.WriteAndRead(argin)
                # self.debug_stream("argout: " + argout)
                if self.get_state() == PyTango.DevState.FAULT:
                    self.set_state(PyTango.DevState.ON)
                if len(argout) == 0:
                    self.status_string = "Cmd failed, no reply"
                    return 0
                elif argout.startswith(resp_ok):
                    self.status_string = "Cmd "+argin+" Successful"
                    return 1
                elif argout.startswith(resp_not_ok):
                    self.status_string = "Cmd "+argin+" Refused"
                    return 0
                else:
                    self.status_string = "Cmd "+argin+" Successful"
                    return argout
                self.debug_stream("Status: " + self.status_string)
            except Exception, e:
                self.error_stream("Socket connection raised exception.")
                attempts += 1
                self.status_string = "Socket Failed:"+str(e)
                time.sleep(0.2)

        # Socket problem persists, set state to FAULT
        self.set_state(PyTango.DevState.FAULT)
        return 0


#------------------------------------------------------------------
#    Write socket utility
#        Just write the command to the socket (max speed).
#------------------------------------------------------------------
    def socket_write(self, argin):
        # print "In ", self.get_name(), "::socket_write()"

        # Flush the socket buffer
        # self.socket.Read()

        argin = str(argin)
        try:
            self.socket.Write(argin)
            # self.debug_stream("argout: " + argout)
            if self.get_state() == PyTango.DevState.FAULT:
                self.set_state(PyTango.DevState.ON)
            return 1
        except Exception, e:
            self.error_stream("Socket connection raised exception.")
            self.status_string = "Socket Failed:"+str(e)
            self.set_state(PyTango.DevState.FAULT)
            return 0



#------------------------------------------------------------------
#    Read Max Voltage from the device utility method
#    Parse max voltage from the device. Based on the firmware version
#    use MTYPE or PARAM:A command. From O.5.4 on, use MTYPE.
#    "Max" voltage can be negative - set as min in such a case.
#    Returns 1 if success, 0 if it fails.
#------------------------------------------------------------------
    def read_maxV_from_device(self):
        # print "In ", self.get_name(), "::read_maxV_from_device()"
        cmd_str = "VER:?"
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            self.debug_stream("Incorrect request, check socket status.")
            return 0
        elif resp.startswith(cmd_str[:-1]):
            version_string = resp[len(cmd_str)-1:]
            version = version_string.split('.')
            if int(version[1]) >= 5 and int(version[2][0])>=4:
                cmd_str = "MTYPE:?"
                sep_chr = ":"
            else:
                cmd_str = "PARAM:A:?"
                sep_chr = ","
            resp = self.socket_write_and_read(cmd_str)
            if resp == 0:
                self.debug_stream("Incorrect request, check socket status.")
                return 0
            elif resp.startswith(cmd_str[:-1]):
                voltage = resp.split(sep_chr)[-1]
                if float(voltage) > 0.0:
                    self.max_voltage = float(voltage)
                else:
                    self.min_voltage = float(voltage)
                self.debug_stream("min_voltage: " + str(self.min_voltage))
                self.debug_stream("max_voltage: " + str(self.max_voltage))
                return 1


#------------------------------------------------------------------
#    Read all the calibration parameters from the device
#------------------------------------------------------------------
    def read_calibration_parameters(self):
        # print "In ", self.get_name(), "::read_calibration_parameters_from_device()"
        self.debug_stream("Reading config params from the device...")
        # 1. Read the coefficients of the calibration linear function from the device
        cmd_str = "DACP:?"
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            self.debug_stream("Incorrect request, check socket status.")
            return 0
        elif resp.startswith(cmd_str[:-1]):
            val = resp[len(cmd_str)-1:]
            val_a, val_b = val.split(',')
            self.calib_slope = float(val_a)
            self.calib_offset = float(val_b)
        else:
            self.debug_stream("Unexpected reply " + str(resp))
            return 0

        # 2. Read the size of the calibration table from the device
        cmd_str = "DACCALN:?"
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            self.error_stream("Incorrect request, check socket status.")
            return 0

        elif resp.startswith(cmd_str[:-1]):
            val = resp[len(cmd_str)-1:]
            self.calib_points = int(val)
        else:
            self.error_stream("Unexpected reply " + str(resp))
            return 0

        # 3. Read the the calibration table from the device
        if self.calib_points == 0:
            self.error_stream("Table size unknown, cannot procede.")
            return 0
        for i in range(self.calib_points):
            cmd_str = "DACCAL:" + str(i) + ",?"
            resp = self.socket_write_and_read(cmd_str)
            if resp == 0:
                self.debug_stream("Incorrect request, check socket status.")
                return 0
            elif resp.startswith(cmd_str[:-1]):
                val = resp[len(cmd_str)-1:]
                self.calib_table.append(int(val))
            else:
                self.debug_stream("Unexpected reply " + str(resp))
                return 0
        self.debug_stream("Calib_slope: " + str(self.calib_slope))
        self.debug_stream("Calib_offset: " + str(self.calib_offset))
        self.debug_stream("Calib_points: " + str(self.calib_points))
        return 1


#------------------------------------------------------------------
#    Set the real voltage on the device to VoltageOffset value 
#    if the real voltage is 0 at startup.
#    Returns 1 if success, 0 if it fails.
#------------------------------------------------------------------
    def initialize_to_offset(self):
        self.debug_stream("VoltageOffset = " + str(self.VoltageOffset))
        if not self.min_voltage <= self.VoltageOffset <= self.max_voltage:
            self.error_stream("VoltageOffset value out of range")
            return 0
        if self.VoltageOffset == 0.0:
            # VoltageOffset not set so ingore the rest
            self.debug_stream("VoltageOffset is 0.")
            return 1
        cmd_str = "HVR:?"
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            self.debug_stream("Incorrect request, check socket status.")
            return 0
        elif resp.startswith(cmd_str[:-1]):
            val = resp[len(cmd_str)-1:]
            if 0.025*self.min_voltage < float(val) < 0.025*self.max_voltage:
                self.debug_stream("Woke up to 0, set to VoltageOffset")
                # Consider this as 0.
                # Set voltage to VoltageOffset.
                self.SetToOffset()
            else:
                self.debug_stream("Device voltage is set to "+str(val))
        return 1


#==================================================================
#
#    FlexHV read/write attribute methods
#
#==================================================================

#------------------------------------------------------------------
#    Read VoltageRange attribute
#------------------------------------------------------------------
    def read_VoltageRange(self, attr):
        voltage_range = "("+str(self.min_voltage) +"V, "+str(self.max_voltage)+"V)"
        attr.set_value(voltage_range)

#------------------------------------------------------------------
#    Read DisplayText attribute
#        Fake: returns last value set, when available.
#------------------------------------------------------------------
    def read_DisplayText(self, attr):
        attr.set_value(self.display_txt)

#------------------------------------------------------------------
#    Write DisplayText attribute
#------------------------------------------------------------------
    def write_DisplayText(self, attr):
        data = attr.get_write_value()
        if len(data)>12:
            raise Exception("Max 12 char allowed.")
        else:
            self.display_txt = data.upper()
            cmd_str = "DISP:"+data.upper()
            if not self.socket_write_and_read(cmd_str):
                raise Exception("Incorrect request: "+self.status_string)

#------------------------------------------------------------------
#    Read CalibratedVoltageOutput attribute
#------------------------------------------------------------------
    def read_CalibratedVoltageOutput(self, attr):
        cmd_str = "HVR:?"
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            raise Exception("Incorrect request: "+self.status_string)
        elif resp.startswith(cmd_str[:-1]):
            val = resp[len(cmd_str)-1:]
            attr.set_value( float(val) )

#------------------------------------------------------------------
#    Write CalibratedVoltageOutput attribute
#------------------------------------------------------------------
    def write_CalibratedVoltageOutput(self, attr):
        data = attr.get_write_value()
        if data < self.min_voltage or data > self.max_voltage:
            self.status_string = "Failed attempt to set voltage."
            raise Exception("Voltage value out of range!")
        else:
            cmd_str = "HVS:"+str(data)
            if not self.socket_write_and_read(cmd_str):
                raise Exception("Incorrect request: "+self.status_string)

#------------------------------------------------------------------
#    Read VoltageOutput attribute
#------------------------------------------------------------------
    def read_VoltageOutput(self, attr):
        cmd_str = "HVR:?"
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            raise Exception("Incorrect request: "+self.status_string)
        elif resp.startswith(cmd_str[:-1]):
            val = resp[len(cmd_str)-1:]
            attr.set_value( float(val) - self.VoltageOffset )

#------------------------------------------------------------------
#    Write VoltageOutput attribute
#------------------------------------------------------------------
    def write_VoltageOutput(self, attr):
        data = attr.get_write_value() + self.VoltageOffset
        if not self.min_voltage <= data <= self.max_voltage:
            self.status_string = "Failed attempt to set voltage."
            raise Exception("Voltage value out of range!")
        else:
            cmd_str = "HVS:"+str(data)
            if not self.socket_write(cmd_str):
                raise Exception("Incorrect request: "+self.status_string)

#------------------------------------------------------------------
#    Read UncalibratedVoltageOutput attribute
#------------------------------------------------------------------
    def read_UncalibratedVoltageOutput(self, attr):
        cmd_str = "HVRB:?"
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            raise Exception("Incorrect request: "+self.status_string)
        elif resp.startswith(cmd_str[:-1]):
            val = resp[len(cmd_str)-1:]
            attr.set_value( int(val) )

#------------------------------------------------------------------
#    Write UncalibratedVoltageOutput attribute
#------------------------------------------------------------------
    def write_UncalibratedVoltageOutput(self, attr):
        data = attr.get_write_value()
        cmd_str = "HVSB:"+str(data)
        if not self.socket_write_and_read(cmd_str):
            raise Exception("Incorrect request: "+self.status_string)

#------------------------------------------------------------------
#    Read CalibratedCurrent attribute
#------------------------------------------------------------------
    def read_CalibratedCurrentOutput(self, attr):
        cmd_str = "HVI:?"
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            raise Exception("Incorrect request: "+self.status_string)
        elif resp.startswith(cmd_str[:-1]):
            val = resp[len(cmd_str)-1:]
            attr.set_value( float(val) )

#------------------------------------------------------------------
#    Read UncalibratedCurrent attribute
#------------------------------------------------------------------
    def read_UncalibratedCurrentOutput(self, attr):
        cmd_str = "HVIB:?"
        resp = self.socket_write_and_read(cmd_str)
        if resp == 0:
            raise Exception("Incorrect request: "+self.status_string)
        elif resp.startswith(cmd_str[:-1]):
            val = resp[len(cmd_str)-1:]
            attr.set_value( int(val) )

#==================================================================
#
#    FlexHV command methods
#
#==================================================================

#------------------------------------------------------------------
#    SetToZero command:
#
#    Description: Software reset of the Voltage
#------------------------------------------------------------------
    def SetToZero(self):
        # print "In ", self.get_name(), "::SetToZero()"
        cmd_str = "HVS:0.0"
        if self.socket_write_and_read(cmd_str):
            self.set_state(PyTango.DevState.ON)
            return "Done"
        else:
            raise Exception("Incorrect request: "+self.status_string)


#------------------------------------------------------------------
#    SetToOffset command:
#
#    Description: Software reset of the Voltage
#------------------------------------------------------------------
    def SetToOffset(self):
        # print "In ", self.get_name(), "::SetToOffset()"
        cmd_str = "HVS:"+str(self.VoltageOffset)
        if self.socket_write_and_read(cmd_str):
            self.set_state(PyTango.DevState.ON)
            return "Done"
        else:
            raise Exception("Incorrect request: "+self.status_string)


#------------------------------------------------------------------
#    ResetFlexDevice command:
#
#    Description: Software reset of the physical device
#------------------------------------------------------------------
    def ResetFlexDevice(self):
        # print "In ", self.get_name(), "::ResetFlexDevice()"
        cmd_str = "*RST"
        if self.socket_write_and_read(cmd_str):
            self.set_state(PyTango.DevState.ON)
            return "Done"
        else:
            raise Exception("Incorrect request: "+self.status_string)

#------------------------------------------------------------------
#    ClearDisplay command:
#
#    Description: Clear the label on the device's display
#------------------------------------------------------------------
    def ClearDisplay(self):
        # print "In ", self.get_name(), "::ClearDisplay()"
        cmd_str = "DISP:CLR"
        if self.socket_write_and_read(cmd_str):
            self.display_txt = ''
            return "Done"
        else:
            #raise Exception("Incorrect request: "+self.status_string)
            self.error_stream("Clear display failed.")
            return "Error"

#------------------------------------------------------------------
#    SetVoltage command:
#
#    Description: Set the voltage to the nearest value possible
#            with the 16 bit converter and return the set value.
#------------------------------------------------------------------
    def SetVoltage(self, voltage):
        # print "In ", self.get_name(), "::SetVoltage(voltage)"
        res = 65535 # 16 bit
        dgt = int(round((voltage + self.VoltageOffset - self.calib_offset)/self.calib_slope))
        seg = int(dgt*self.calib_points/float(res))
        # self.debug_stream("Calib segment: " + str(seg))
        d_corr = dgt + self.calib_table[seg]
        if d_corr < 0:
            d_corr = 0
        if d_corr > res:
            d_corr = res
        real_voltage = dgt * self.calib_slope + self.calib_offset - self.VoltageOffset
        cmd_str = "HVSB:"+str(d_corr)
        if self.socket_write(cmd_str):
            return real_voltage
        else:
            raise Exception("Incorrect request: "+self.status_string)
        

#==================================================================
#
#    FlexHVClass class definition
#
#==================================================================
class FlexHVClass(PyTango.DeviceClass):

    #    Class Properties
    class_property_list = {
        }

    #    Device Properties
    device_property_list = {
        'SocketDevice':
            [PyTango.DevString,
            "Socket TangoDevice for communication with the Flex HV PS.",
            ["/some/tango/device"]],
        'VoltageOffset':
            [PyTango.DevDouble,
            "Voltage Offset in V.",
            [0.0]],
        }

    #    Command definitions
    cmd_list = {
        'SetToZero':
            [[PyTango.DevVoid, "Sets Voltage 0.0"],
            [PyTango.DevString, "ACK/NAK"]],
        'SetToOffset':
            [[PyTango.DevVoid, "Sets Voltage to VoltageOffset property value"],
            [PyTango.DevString, "ACK/NAK"]],
        'ResetFlexDevice':
            [[PyTango.DevVoid, "Software reset on the FLEX device: return to PowerON state."],
            [PyTango.DevString, "ACK/NAK"]],
        'ClearDisplay':
            [[PyTango.DevVoid, "Reset OLED Display."],
            [PyTango.DevString, "ACK/NAK"]],
        'SetVoltage':
            [[PyTango.DevDouble, "Set the desired voltage."],
            [PyTango.DevDouble, "Calculated actual voltage."]],        
        }

    #    Attribute definitions
    attr_list = {
        'VoltageRange':
            [[PyTango.DevString,
            PyTango.SCALAR,
            PyTango.READ],
            {
                'description':"HV Module max voltage",
            }],
        'DisplayText':
            [[PyTango.DevString,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'description':"Label that appears on display, max 12 upper key chars, CLR to clear.",
            }],
        'CalibratedVoltageOutput':
            [[PyTango.DevDouble,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'description':"Actual voltage in V, [0, MaxVoltage] or [MinVoltage, 0].",
                'unit':'V',
                'Memorized': "false",
            }],
        'VoltageOutput':
            [[PyTango.DevDouble,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'description':"Voltage in V considering the VoltageOffset.",
                'unit':'V',
                'Memorized': "false",
            }],
        'UncalibratedVoltageOutput':
            [[PyTango.DevShort,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'description':"Binary RAW voltage value in range 0-65535.",
            }],
        'CalibratedCurrentOutput':
            [[PyTango.DevDouble,
            PyTango.SCALAR,
            PyTango.READ],
            {
                'description':"Current value in uA",
                'unit':'uA',
            }],
        'UncalibratedCurrentOutput':
            [[PyTango.DevDouble,
            PyTango.SCALAR,
            PyTango.READ],
            {
                'description':"Binary RAW current value in range 0-4095.",
            }],
        }


#------------------------------------------------------------------
#    FlexHVClass Constructor
#------------------------------------------------------------------
    def __init__(self, name):
        print 'In FlexHVClass constructor'
        PyTango.DeviceClass.__init__(self, name)
        self.set_type(name)

#==================================================================
#
#    FlexHV class main method
#
#==================================================================
if __name__ == '__main__':
    try:
        py = PyTango.Util(sys.argv)
        py.add_TgClass(FlexHVClass,FlexHV,'FlexHV')

        U = PyTango.Util.instance()
        U.server_init()
        U.server_run()

    except PyTango.DevFailed,e:
        print '-------> Received a DevFailed exception:',e
    except Exception,e:
        print '-------> An unforeseen exception occured....',e

